/*      lalign.c
	protein driver for linear sequence comparison method

	March, 1995 - increased space for initseq() for long long gaps,
	changed default gap penalty to -14, -4 for protein.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*#include <ctype.h>*/

char *refstr="\nPlease cite:\n X. Huang and W. Miller (1991) Adv. Appl. Math. 12:373-381\n";
char *verstr="version 2.0u66 September 1998";
char *progstr="LALIGN";

#ifdef __MWERKS__
#include <Types.h>
#include <StandardFile.h>
StandardFileReply freply;
Point wpos;
pascal	OSErr SetDefault(FSSpec * f_new, FSSpec *f_old);
pascal	OSErr RestoreDefault(FSSpec *f_info);

int tval;
char prompt[256];
#include <sioux.h>
#define getenv mgetenv
#endif

#ifndef BIGMEM
#define MAXTST 2000	/* longest test sequence */
#define MAXLIB 2000
#define MAXDIAG (MAXTST+MAXLIB)
#define QFILE_SIZE 40
#else
#define MAXTST 20000
#define MAXLIB 120000
#define MAXDIAG (MAXTST+MAXLIB)
#define QFILE_SIZE 256
#endif

FILE *outfd;		/* fd for output file */

extern int init_stats(char *pam_type, int gdelval, int ggapval);

/* globals for matching */

long lmark;		/* position in library file from ftell() */

char libstr[21];	/* partial title from library sequence */
char name0[11], name1[11];	/* for labeling output */

char *aa0=NULL, *aa1=NULL;	/* amino acid sequence data */
char *seqc0, *seqc1;	/* aligned sequences */
long sq0off=1, sq1off=1;

int dnaseq, lcont;
int have_stats=0;
int bktup, bkfact, scfact, bestoff, bestscale, histint, bestmax;

int nseq=2;		/* same sequence twice */
int maxn;		/* max space for lib sequence */
int n0, n1;	/* length of aa0, length of aa1, n0+n1,	diagonal offset */
long loffset = 0l;		/* offset into sequence */

/*  the following are defaults for values that are read by
    pam.c from *.mat if SMATRIX is defined */

int nshow;
char rline[20],sline[20];
char resfile[QFILE_SIZE];

/* output options */
int showall,markx, llen;

char ttitle[60], ltitle[60];

int smark[4] = {-10000,-10000,-10000,-10000};
int revflg = 0;
char *compstr="\0";

int min0,min1,max0,max1,mins;
#ifdef TPLOT
char lvstr[40];
#endif

extern int optind;
char *libenv, *aaenv, *smptr;
char smstr[QFILE_SIZE];
char pam_type[10];

#ifdef __MWERKS__
/* short ouvRef, q0vRef, q1vRef; */
FSSpec ouSpec, q0Spec, q1Spec;
OSErr error;
#define IntroDID 400
#endif

#ifdef TPLOT
char *iprompt0=" PLALIGN finds the best local alignments between two sequences\n";
#else
char *iprompt0=" LALIGN finds the best local alignments between two sequences\n";
#endif
char *iprompt1=" first sequence name: ";
char *iprompt2=" second sequence name: ";

#include "upam.gbl"		/* includes pam array */

main(argc, argv)
     int argc; char **argv;
{
  char tname[QFILE_SIZE], lname[QFILE_SIZE], qline[QFILE_SIZE];
  int K;
  int itemp, iln, nln, i;
  char *getenv(), *cptr, *bp;
  float percent;


#ifdef __MWERKS__
#ifdef TPLOT
	InitGraf(&qd.thePort);
#endif
  SIOUXSettings.asktosaveonclose=TRUE;
  SIOUXSettings.showstatusline=FALSE;
  SIOUXSettings.autocloseonquit=FALSE;

  argc = ccommand(&argv);
  if (GetResource('DLOG',IntroDID)==(Handle)0 && OpenResFile("\pFASTA.rsrc")<0) {
    SysBeep(100); fprintf(stderr," WARNING FASTA.rsrc file could not be found\n");
  }
  InitEvent();
  error=HGetVol(NULL,&ouSpec.vRefNum, &ouSpec.parID);
  if (error != noErr) {
  	fprintf(stderr," cannot get current directory\n");
  	exit(1);
  	}
  /* GetVol((StringPtr)prompt,&ouvRef); */
  wpos.h=50; wpos.v=100;
#endif

  initenv(argc,argv);

  if ((aa0=calloc((size_t)MAXTST+MAXLIB,sizeof(char)))==0) {
    fprintf(stderr," cannot allocate sequence array\n");
    exit(1);
  }
  maxn = MAXTST+MAXLIB;

  if (argc-optind < 3) {
#ifndef __MWERKS__
    fputs(iprompt0,stderr);
    fprintf(stdout," %s%s\n",verstr,refstr);

  l1: fputs(iprompt1,stderr);
    fflush(stdout);
    if (fgets(tname,sizeof(tname),stdin)==NULL) exit(0);
    if (tname[strlen(tname)-1]=='\n') tname[strlen(tname)-1]='\0';
    if (tname[0]=='\0') goto l1;
#else
    NIntroDlog(IntroDID,iprompt0,verstr,refstr,"\0");
	  
  l1:	FileDlog(iprompt1,&freply);
    if (freply.sfGood==TRUE) {
      PtoCstr(freply.sfFile.name);
      strcpy(tname,(char *)freply.sfFile.name);
	  q0Spec.vRefNum = freply.sfFile.vRefNum;
	  q0Spec.parID = freply.sfFile.parID;
	  HSetVol(NULL,q0Spec.vRefNum,q0Spec.parID);
    }
    else exit(0);
#endif
    if ((n0=getseq(tname,aa0,maxn,&dnaseq))==0) {
      fprintf(stderr," %s : %s sequence not found\n",tname,sqtype);
      goto l1;
    }

    resetp(dnaseq);
    dnaseq = -1;
    if (revflg) {
      if (sqtype[0]=='D') revcomp(aa0,n0);
      else {
	fprintf(stderr," can only reverse complement DNA\n");
	compstr="\0"; revflg = 0;
      }
    }
    

#ifndef __MWERKS__
  l2:	fputs(iprompt2,stderr);
    fflush(stderr);
    if (fgets(lname,sizeof(lname),stdin)==NULL) exit(1);
    if (lname[strlen(lname)-1]=='\n') lname[strlen(lname)-1]='\0';
    if (*lname==0) goto l2;
#else
  l2:	FileDlog(iprompt2,&freply);
    if (freply.sfGood==TRUE) {
      PtoCstr((StringPtr)freply.sfFile.name);
      strcpy(lname,(char *)freply.sfFile.name);
	  q1Spec.vRefNum = freply.sfFile.vRefNum;
	  q1Spec.parID = freply.sfFile.parID;
	  HSetVol(NULL,q1Spec.vRefNum,q1Spec.parID);
    }
    else exit(0);
#endif

  l3:	fprintf(stderr," number of local alignments [10]: ");
    fflush(stderr);
    if (fgets(rline,sizeof(rline),stdin)==NULL) K=10;
    else {
      if (rline[0]!='\0' && rline[0]!='\n') sscanf(rline,"%d",&K);
      else K=10;
      if (K<1) K=10;
    }
  }
  else {
    fputs(iprompt0,stderr);
    fprintf(stderr," %s%s\n",verstr,refstr);
    strncpy(tname,argv[optind+1],sizeof(tname));
    if ((n0=getseq(tname,aa0,maxn,&dnaseq))==0) {
      fprintf(stderr," %s : %s sequence not found\n",tname,sqtype);
      exit(1);
    }
    resetp(dnaseq);
    dnaseq = -1;
    if (revflg) {
      if (dnaseq) revcomp(aa0,n0);
      else {
	fprintf(stderr," cannot reverse complement protein sequence\n");
	compstr="\0"; revflg = 0;
      }
    }
    strncpy(lname,argv[optind+2],sizeof(tname));
    if (argc > optind+3) sscanf(argv[optind+3],"%d",&K);
    else K=10;
  }

  if (strcmp(tname,lname)==0 && !revflg && (lname[0]!='-' && lname[0]!='@'))
      nseq=1;
  else nseq=2;

  strncpy(name0,tname,6);
  gettitle(tname,ttitle,50);
  if (strlen(ttitle)>0)
    if (*ttitle=='>') strncpy(name0,&ttitle[1],6);
    else strncpy(name0,ttitle,6);
  else
    strncpy(name0,tname,6);
  name0[6]='\0';
  if (revflg) name0[5]='-';

  if ((bp=strchr(name0,' '))!=NULL) *bp='\0';

  outfd = stdout;
  if (resfile[0]!='\0')
    if ((outfd=fopen(resfile,"w"))==NULL) outfd = stdout;
#if defined(TPLOT) && defined(__MWERKS__)
    else {
      SIOUXSettings.asktosaveonclose=FALSE;
      SIOUXSettings.autocloseonquit=TRUE;
    }
#endif
		

  aa1 = aa0 + n0 + 2;
  maxn -= n0 + 3;

  if (nseq==2) {
    n1=getseq(lname,aa1,maxn,&dnaseq);
    if (n1 <= 1) {
      fprintf(stderr," cannot read %s\n",lname);
      exit(1);
    }
    gettitle(lname,ltitle,50);
    if (strlen(ltitle)>0)
      if (*ltitle=='>') strncpy(name1,&ltitle[1],6);
      else strncpy(name1,ltitle,6);
    else strncpy(name1,lname,6);
    name1[6]='\0';
    if ((bp=strchr(name1,' '))!=NULL) *bp='\0';

    if (n1 == n0) {
      for (i=0; i<n0; i++)
	if (aa0[i]!=aa1[i]) break;
      if (i==n0) nseq = 1;
    }
  }
  else {
    aa1 = aa0;
    n1 = n0;
    strncpy(name1,name0,6);
    strncpy(ltitle,ttitle,sizeof(ltitle));
  }

#ifdef __MWERKS__
	 HSetVol(NULL,ouSpec.vRefNum,ouSpec.parID);
/*   SetVol("\p\0",ouvRef); */
#endif

  if (markx < 10) {
#ifndef TPLOT
    fprintf(outfd," Comparison of:\n(A) %-10s%s %-50s - %d %s\n",
	    tname,compstr,ttitle,n0,sqnam);
    fprintf(outfd,"(B) %-10s %-50s - %d %s\n",lname,ltitle,n1,sqnam);
    fprintf(outfd," using matrix file: %s, gap penalties: %d/%d\n",
	    smptr,gdelval,ggapval);
#endif
  }
  else if (markx==10) {
    fprintf(outfd,"\n>>>%s%s, %d %s vs %s, %d %s\n",
	    tname, compstr, n0, sqnam, lname, n1, sqnam);
    fprintf(outfd,"; pg_name: %s\n",progstr);
    fprintf(outfd,"; pg_ver: %s\n",verstr);
    fprintf(outfd,"; pg_matrix: %s\n",smptr);
    fprintf(outfd,"; pg_gap-pen: %d %d\n",gdelval,ggapval);
  }

  initseq(min(n0,n1)*2);

  initpam2();			/* convert 1-d pam to 2-d pam2 */

  have_stats = init_stats(smptr,gdelval,ggapval);

#ifdef TPLOT
  openplt((long)n0,(long)n1);
  if (nseq==1) drawdiag((long)n0,(long)n1);
  SIM(aa0-1,aa1-1,n0,n1,K,pam2,-(gdelval-ggapval),-ggapval,nseq);
  closeplt();
#else
  SIM(aa0-1,aa1-1,n0,n1,K,pam2,-(gdelval-ggapval),-ggapval,nseq);
#endif
  exit(0);
}

extern int *sascii, nascii[], aascii[];

initenv(argc,argv)
     int argc;
     char **argv;
{
  char *cptr, *getenv();
  int copt, getopt();
  extern char *optarg;

  libenv="\0";
  aaenv="\0";

  sascii = aascii;
  pam = abl50;
  gdelval = -14;
  ggapval = -4;
  strncpy(smstr,"BL50",sizeof(smstr));
  smptr=smstr;
  sq = aa;
  hsq = haa;
  nsq = naa;
  dnaseq = 0;

  showall = 1;

  if ((cptr=getenv("LINLEN"))!=NULL) sscanf(cptr,"%d",&llen);
  else llen = 60;
  if (llen>=200) llen=200-1;
  markx=0;
  if ((cptr=getenv("MARKX"))==NULL) markx=0;
  else sscanf(cptr,"%d",&markx);

  while ((copt=getopt(argc,argv,"f:g:im:nO:Qqs:w:v:x:"))!=EOF)
    switch(copt) {
    case 'w': sscanf(optarg,"%d",&llen); break;
    case 'f': sscanf(optarg,"%d",&gdelval); 
      del_set=1;
      break;
    case 'g': sscanf(optarg,"%d",&ggapval); 
      gap_set=1;
      break;
    case 'i': revflg = 1; compstr=" (rev-comp)"; break;
    case 'm': sscanf(optarg,"%d",&markx); break;
    case 'n': dnaseq=1;
      sascii = nascii;
      sq = nt;
      nsq = nnt;
      hsq = hnt;
      pam = npam;
      strcpy(sqnam,"nt");
      strcpy(sqtype,"DNA");
      resetp(dnaseq);
      break;
    case 'O': strncpy(resfile,optarg,sizeof(resfile));
      break;
    case 'q':
    case 'Q':
      break;
    case 's': strncpy(smstr,optarg,sizeof(smstr));
      smptr = smstr;
      if (strcmp(smstr,"BL62")==0) {
	pam = abl62;
	dnaseq= -1;
      }
      else if (strcmp(smstr,"P250")==0) {
	pam = apam250;
	dnaseq= -1;
      }
      else if (initpam(smptr)) {
	dnaseq= -1;
      }
      break;
    case 'x': sscanf(optarg,"%ld %ld",&sq0off,&sq1off);
      break;
#ifdef TPLOT
    case 'v': strncpy(lvstr,optarg,sizeof(lvstr));
      break;
#endif
    default : fprintf(stderr," illegal option -%c\n",copt);
    }
  optind--;
  if (dnaseq>=0) {
    if ((smptr=getenv("SMATRIX"))!=NULL && initpam(smptr)) {
      dnaseq = -1;
    }
    else smptr=smstr;
  }
}

resetp(dnaseq)
     int dnaseq;
{
  if (dnaseq==1) {
    fprintf(stderr," resetting to DNA matrix\n");
    pam = npam;
    strncpy(smstr,"DNA",sizeof(smstr));
    smptr = smstr;
    if (!del_set) gdelval = -16;
  }
}

int match, mismh;

initpam2()
{
	int i, j, k, tmp;

	match = -1000; mismh = 1000;
	k=0;
	for (i=0; i<nsq; i++)
		for (j=0; j<=i; j++) {
			tmp=pam2[j][i] = pam2[i][j] = pam[k++];
			if (tmp>match) match=tmp;
			if (tmp<mismh) mismh=tmp;
		      }
	}

int smin0, smin1, smins;	/* set bounds for discons */

calcons(aa0,n0,aa1,n1,res,nc,nident)
     char *aa0, *aa1;
     int n0, n1;
     int *res;
     int *nc;
     int *nident;
{
  int i0, i1;
  int op, nid, lenc, nd, ns, itmp;
  char *sp0, *sp1;
  int *rp;
  
  /* first fill in the ends */
  min0--; min1--;

  smin0 = min0;
  smin1 = min1;
  smins = mins = 0;

/* now get the middle */

  sp0 = seqc0+mins;
  sp1 = seqc1+mins;
  rp = res;
  lenc = nid = op = 0;
  i0 = min0;
  i1 = min1;
  
  while (i0 < max0 || i1 < max1) {
    if (op == 0 && *rp == 0) {
      op = *rp++;
      *sp0 = sq[aa0[i0++]];
      *sp1 = sq[aa1[i1++]];
      lenc++;
      if (*sp0 == *sp1) nid++;
      else if ((dnaseq==1) && (*sp0=='T' && *sp1=='U') ||
	       (*sp0=='U' && *sp1=='T')) nid++;
      sp0++; sp1++;
    }
    else {
      if (op==0) op = *rp++;
      if (op>0) {
	*sp0++ = '-';
	*sp1++ = sq[aa1[i1++]];
	op--;
	lenc++;
      }
      else {
	*sp0++ = sq[aa0[i0++]];
	*sp1++ = '-';
	op++;
	lenc++;
      }
    }
  }

  *nident = nid;
  *nc = lenc;
/*	now we have the middle, get the right end */
  nd = 0;
  return mins+lenc+nd;
}

initseq(seqsiz)		/* initialize arrays */
	int seqsiz;
{
	seqc0=calloc((size_t)seqsiz,sizeof(char));
	seqc1=calloc((size_t)seqsiz,sizeof(char));
	if (seqc0==NULL || seqc1==NULL)
		{fprintf(stderr,"cannot allocate consensus arrays %d\n",seqsiz);
		 exit(1);}
	}

freeseq()
{
	free(seqc0); free(seqc1);
	}

